"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import bcrypt
import random
import string


def generate_random(length: int) -> str:
    """
    Generates a random byte string consisting of alphanumeric characters
    Thanks @ Albert (https://stackoverflow.com/users/281021/albert)
    https://stackoverflow.com/questions/2257441
    :param length: The length of the string to generate
    :return: The generated random byte string
    """
    return bytes(
        "".join(random.choice(string.ascii_uppercase + string.digits)
                for _ in range(length)),
        "utf-8"
    ).decode("utf-8")


def generate_hash(password: str) -> str:
    """
    Salts and hashes a password to generate a hash for storage in a database
    :param password: The password to hash
    :return: The hash of the password
    """
    password_bytes = bytes(password, "utf-8")
    return bcrypt.hashpw(password_bytes, bcrypt.gensalt()).decode("utf-8")


def verify_password(password: str, hashed: str):
    """
    Verifies that a password matches a given hash
    :param password: The password to verify
    :param hashed: The hash to verify the password against
    :return: True if the password matches, otherwise False
    """
    password_bytes = bytes(password, "utf-8")
    hashed_bytes = bytes(hashed, "utf-8")

    try:
        return bcrypt.checkpw(password_bytes, hashed_bytes)
    except ValueError:
        return False



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""


from typing import Dict, Any


def jsonify_models(data: Dict[str, Any], deep: bool = True) -> Dict[str, Any]:
    """
    Serializes a dictionary and calls the __json__() method on all
    objects that have one.
    :param data: The data to serialize
    :param deep: Indicates whether or not child models are included.
                 If False, only IDs will be included.
    :return: The serialized data
    """
    to_jsonify = {}
    for key, val in data.items():
        to_jsonify[key] = _serialize(val, deep)
    return to_jsonify


def _serialize(obj: Any, deep: bool) -> Any:
    """
    Serializes an object, so that it can be jsonified without issues
    :param obj: The object to serialize
    :param deep: Indicates whether or not child models are included.
                 If False, only IDs will be included.
    :return: The serialized object
    """
    if isinstance(obj, dict):
        serialized_dict = {}
        for key, val in obj.items():
            serialized_dict[key] = _serialize(val, deep)
        return serialized_dict

    elif isinstance(obj, list):
        serialized_list = []
        for val in obj:
            serialized_list.append(_serialize(val, deep))
        return serialized_list

    elif isinstance(obj, tuple):
        return tuple(_serialize(list(obj), deep))

    elif "__json__" in dir(obj):
        return obj.__json__(deep)
    else:
        return obj



"""LICENSE
Copyright 2020 Hermann Krumrey <hermann@krumreyh.com>

This file is part of fat-ffipd.

fat-ffipd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

fat-ffipd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with fat-ffipd.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Union
from flask import render_template, Blueprint
from werkzeug import Response
from puffotter.flask.Config import Config


def define_blueprint(blueprint_name: str) -> Blueprint:
    """
    Defines the blueprint for this route
    :param blueprint_name: The name of the blueprint
    :return: The blueprint
    """
    blueprint = Blueprint(blueprint_name, __name__)

    @blueprint.route("/", methods=["GET"])
    def index() -> Union[Response, str]:
        """
        The index page
        :return: The response
        """
        return render_template(
            "static/index.html",
            **Config.TEMPLATE_EXTRAS["index"]()
        )

    @blueprint.route("/about", methods=["GET"])
    def about() -> Union[Response, str]:
        """
        The about page/"Impressum" for the website
        :return: The response
        """
        return render_template(
            "static/about.html",
            **Config.TEMPLATE_EXTRAS["about"]()
        )

    @blueprint.route("/privacy", methods=["GET"])
    def privacy() -> Union[Response, str]:
        """
        Page containing a privacy disclaimer
        :return: The response
        """
        return render_template(
            "static/privacy.html",
            **Config.TEMPLATE_EXTRAS["privacy"]()
        )

    return blueprint



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from unittest.mock import patch
from puffotter.flask.test.TestFramework import _TestFramework


class TestErrorHandling(_TestFramework):
    """
    Class that tests the flask error handling
    """

    def test_exception(self):
        """
        Tests if unexpected exceptions are caught correctly
        :return: None
        """

        def render_template(*_, **__):
            print(1/0)
        with patch("puffotter.flask.routes.static.render_template",
                   render_template):
            resp = self.client.get("/", follow_redirects=True)
            self.assertTrue(b"The server encountered an internal" in resp.data)

    def test_404(self):
        """
        Tests if a 404 error is handled correctly
        :return: None
        """
        resp = self.client.get("/baba", follow_redirects=True)
        print(resp.data)
        self.assertTrue(b"404 Not Found" in resp.data)



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import json
from unittest.mock import patch
from flask.blueprints import Blueprint
from puffotter.flask.routes.decorators import api
from puffotter.flask.test.TestFramework import _TestFramework


class TestApiCalls(_TestFramework):
    """
    Class that tests various API calls
    """

    def test_unauthorized_call(self):
        """
        Tests and unauthorized API call
        :return: None
        """
        api_headers = self.generate_api_key_headers("1:100")
        resp = self.client.get(
            "/api/v1/authorize", headers=api_headers, json={}
        )
        self.assertEqual(resp.status_code, 401)
        data = json.loads(resp.data.decode("utf-8"))
        self.assertEqual(data["status"], "error")
        self.assertEqual(data["reason"], "unauthorized")

    def test_non_base64_header(self):
        """
        Tests using a header that's not base64 encoded
        :return: None
        """
        user, _, _ = self.generate_sample_user()
        _, api_key, _ = self.generate_api_key(user)
        resp = self.client.get(
            "/api/v1/authorize",
            headers={"Authorization": "Basic " + api_key},
            json={}
        )
        self.assertEqual(resp.status_code, 401)
        data = json.loads(resp.data.decode("utf-8"))
        self.assertEqual(data["status"], "error")
        self.assertEqual(data["reason"], "unauthorized")

    def test_expired_api_key(self):
        """
        Tests using an expired API key
        :return: None
        """
        user, _, _ = self.generate_sample_user()
        api_key_obj, api_key, _ = self.generate_api_key(user)

        api_key_obj.creation_time = 0
        self.db.session.commit()

        api_headers = self.generate_api_key_headers(api_key)
        resp = self.client.get(
            "/api/v1/authorize", headers=api_headers, json={}
        )
        self.assertEqual(resp.status_code, 401)
        data = json.loads(resp.data.decode("utf-8"))
        self.assertEqual(data["status"], "error")
        self.assertEqual(data["reason"], "unauthorized")

    def test_using_non_json_data(self):
        """
        Tests sending the data as something that's not JSON
        :return: None
        """
        user, password, _ = self.generate_sample_user()
        resp = self.client.post("/api/v1/key", data={
            "username": user.username,
            "password": password
        })
        self.assertEqual(resp.status_code, 400)
        data = json.loads(resp.data.decode("utf-8"))
        self.assertEqual(data["status"], "error")
        self.assertEqual(data["reason"], "not in json format")

    def test_random_exception(self):
        """
        Tests that the API routes catch any Exceptions without issue
        :return: None
        """
        class Mocker:
            @staticmethod
            def get_json():
                print({}["test"])

        with patch("puffotter.flask.routes.api.user_management.request",
                   Mocker):
            user, password, _ = self.generate_sample_user()
            resp = self.client.post("/api/v1/key", json={
                "username": user.username,
                "password": password
            })
            self.assertEqual(resp.status_code, 400)
            data = json.loads(resp.data.decode("utf-8"))
            self.assertEqual(data["status"], "error")
            self.assertEqual(data["reason"], "bad request: KeyError")

    def test_exception_in_api_route(self):
        """
        Tests if an exeption on an API route is correctly wrapped
        in a JSON response
        :return: None
        """
        bp = Blueprint("exceptiontest", __name__)

        @bp.route("/api/exception", methods=["GET"])
        @api
        def exception_route():
            return f"{1/0}"

        self.app.register_blueprint(bp)
        resp = self.client.get("/api/exception")
        self.assertEqual(resp.status_code, 500)



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from puffotter.flask.test.TestFramework import _TestFramework
from puffotter.flask.db.User import User


class TestProfileRoute(_TestFramework):
    """
    Class that tests profile features
    """

    def test_page_get(self):
        """
        Tests getting the page
        :return: None
        """
        user, password, _ = self.generate_sample_user()
        with self.client:
            self.login_user(user, password)
            resp = self.client.get("/profile")
            self.assertTrue(
                b"<!--user_management/profile.html-->" in resp.data
            )
            self.assertTrue(user.username.encode("utf-8") in resp.data)

    def test_changing_password(self):
        """
        Tests changing a password
        :return: None
        """
        user, password, _ = self.generate_sample_user()
        new_pass = "abcdefg"
        with self.client:
            self.login_user(user, password)
            resp = self.client.post("/change_password", follow_redirects=True,
                                    data={
                                        "old_password": password,
                                        "new_password": new_pass,
                                        "password_repeat": new_pass
                                    })
            self.assertTrue(
                b"<!--user_management/profile.html-->" in resp.data
            )
            password_changed = self.config.STRINGS["password_changed"]
            self.assertTrue(password_changed.encode("utf-8") in resp.data)
            self.assertFalse(user.verify_password(password))
            self.assertTrue(user.verify_password(new_pass))

    def test_unsuccessful_password_change(self):
        """
        Tests unsuccessfully changing a password
        :return: None
        """
        user, password, _ = self.generate_sample_user()
        new_pass = "abcdefg"
        base = {
            "old_password": password,
            "new_password": new_pass,
            "password_repeat": new_pass
        }
        with self.client:
            self.login_user(user, password)
            for params in [
                {"new_password": "AAA", "expected": "passwords_do_not_match"},
                {"old_password": "AAA", "expected": "invalid_password"}
            ]:
                data = dict(base)
                data.update(params)

                resp = self.client.post(
                    "/change_password", follow_redirects=True, data=data
                )
                self.assertTrue(
                    b"<!--user_management/profile.html-->" in resp.data
                )
                expected = self.config.STRINGS[params["expected"]]
                password_changed = self.config.STRINGS["password_changed"]
                self.assertTrue(expected.encode("utf-8") in resp.data)
                self.assertFalse(password_changed.encode("utf-8") in resp.data)
                self.assertTrue(user.verify_password(password))
                self.assertFalse(user.verify_password(new_pass))

    def test_user_delete(self):
        """
        Tests deleting a user
        :return: None
        """
        user, password, _ = self.generate_sample_user()
        with self.client:
            self.login_user(user, password)
            self.assertEqual(len(User.query.all()), 1)

            resp = self.client.post("/delete_user", follow_redirects=True,
                                    data={"password": password})
            user_deleted = self.config.STRINGS["user_was_deleted"]

            self.assertTrue(b"<!--static/index.html-->" in resp.data)
            self.assertTrue(user_deleted.encode("utf-8") in resp.data)
            self.assertEqual(len(User.query.all()), 0)

    def test_unsuccessful_user_delete(self):
        """
        Tests unsuccessfully deleting a user
        :return: None
        """
        user, password, _ = self.generate_sample_user()
        with self.client:
            self.login_user(user, password)
            self.assertEqual(len(User.query.all()), 1)
            resp = self.client.post("/delete_user", follow_redirects=True,
                                    data={"password": "AAA"})
            self.assertTrue(
                b"<!--user_management/profile.html-->" in resp.data
            )
            invalid_password = self.config.STRINGS["invalid_password"]
            self.assertTrue(invalid_password.encode("utf-8") in resp.data)
            self.assertEqual(len(User.query.all()), 1)



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import time
from typing import Dict, Any
from puffotter.flask.db.User import User
from puffotter.flask.base import db
from puffotter.flask.Config import Config
from puffotter.flask.db.ModelMixin import ModelMixin
from puffotter.crypto import verify_password


class ApiKey(ModelMixin, db.Model):
    """
    Model that describes the 'api_keys' SQL table
    An ApiKey is used for API access using HTTP basic auth
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes the Model
        :param args: The constructor arguments
        :param kwargs: The constructor keyword arguments
        """
        super().__init__(*args, **kwargs)

    __tablename__ = "api_keys"
    """
    The name of the table
    """

    user_id: int = db.Column(
        db.Integer,
        db.ForeignKey("users.id"),
        nullable=False
    )
    """
    The ID of the user associated with this API key
    """

    user: User = db.relationship("User", back_populates="api_keys")
    """
    The user associated with this API key
    """

    key_hash: str = db.Column(db.String(255), nullable=False)
    """
    The hash of the API key
    """

    creation_time: int = \
        db.Column(db.Integer, nullable=False, default=time.time)
    """
    The time at which this API key was created as a UNIX timestamp
    """

    def has_expired(self) -> bool:
        """
        Checks if the API key has expired.
        API Keys expire after 30 days
        :return: True if the key has expired, False otherwise
        """
        return time.time() - self.creation_time > Config.MAX_API_KEY_AGE

    def verify_key(self, key: str) -> bool:
        """
        Checks if a given key is valid
        :param key: The key to check
        :return: True if the key is valid, False otherwise
        """
        try:
            _id, api_key = key.split(":", 1)
            if int(_id) != self.id:
                return False
            else:
                return verify_password(api_key, self.key_hash)
        except ValueError:
            return False

    def __json__(self, include_children: bool = False) -> Dict[str, Any]:
        """
        Generates a dictionary containing the information of this model
        :param include_children: Specifies if children data models will be
                                 included or if they're limited to IDs
        :return: A dictionary representing the model's values
        """
        data = {
            "id": self.id,
            "user_id": self.user_id,
            "creation_time": self.creation_time
        }
        if include_children:
            data["user"] = self.user.__json__(include_children)
        return data



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from typing import Dict, Any
from bokkichat.entities.Address import Address
from bokkichat.entities.message.TextMessage import TextMessage
from puffotter.flask.base import db
from puffotter.flask.Config import Config
from puffotter.flask.db.ModelMixin import ModelMixin
from puffotter.flask.db.User import User


class TelegramChatId(ModelMixin, db.Model):
    """
    Model that describes the 'telegram_chat_ids' SQL table
    Maps telegram chat ids to users
    """

    def __init__(self, *args, **kwargs):
        """
        Initializes the Model
        :param args: The constructor arguments
        :param kwargs: The constructor keyword arguments
        """
        super().__init__(*args, **kwargs)

    __tablename__ = "telegram_chat_ids"
    """
    The name of the table
    """

    user_id: int = db.Column(
        db.Integer, db.ForeignKey("users.id"),
        nullable=False
    )
    """
    The ID of the user associated with this telegram chat ID
    """

    user: User = db.relationship("User", back_populates="telegram_chat_id")
    """
    The user associated with this telegram chat ID
    """

    chat_id: str = db.Column(db.String(255), nullable=False)
    """
    The telegram chat ID
    """

    def send_message(self, message_text: str):
        """
        Sends a message to the telegram chat
        :param message_text: The message text to send
        :return: None
        """
        address = Address(self.chat_id)
        message = TextMessage(
            Config.TELEGRAM_BOT_CONNECTION.address,
            address,
            message_text
        )
        Config.TELEGRAM_BOT_CONNECTION.send(message)

    def __json__(self, include_children: bool = False) -> Dict[str, Any]:
        """
        Generates a dictionary containing the information of this model
        :param include_children: Specifies if children data models will be
                                 included or if they're limited to IDs
        :return: A dictionary representing the model's values
        """
        data = {
            "id": self.id,
            "user_id": self.user_id,
            "chat_id": self.chat_id
        }
        if include_children:
            data["user"] = self.user.__json__(include_children)
        return data



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import os
import shutil
from typing import Tuple, List, Optional


def listdir(
        path: str,
        no_files: bool = False,
        no_dirs: bool = False,
        no_dot: bool = True
) -> List[Tuple[str, str]]:
    """
    Improves on the standard os.listdir function.
    By default, files and directories starting with `.` are ignored and
    one can disable checking for files or directories.
    Instead of just the filenames, the function returns tuples of filenames
    and relative file paths.
    :param path: The path of which to list the contents
    :param no_files: If set to True, will ignore files
    :param no_dirs: If set to True, will ignore directories
    :param no_dot: If set to True, will ignore files starting with `.`
    :return: A sorted list of the contents of the directory, consisting of
             tuples of the file/directory name and their relative path.
    """
    content = []
    for child in sorted(os.listdir(path)):
        child_path = os.path.join(path, child)

        if no_files and os.path.isfile(child_path):
            continue
        elif no_dirs and os.path.isdir(child_path):
            continue
        elif no_dot and child.startswith("."):
            continue
        else:
            content.append((child, child_path))
    return content


def makedirs(path: str, delete_before: bool = False):
    """
    A more pleasant to use makedirs function.
    Only calls os.makedirs if the directory does not already exist
    :param path: The path to the directory to create
    :param delete_before: If True, deletes the directory beforehand,
                          if it exists
    :return: None
    """
    if delete_before and os.path.isdir(path):
        shutil.rmtree(path)

    if not os.path.isdir(path):
        os.makedirs(path)


def replace_illegal_ntfs_chars(string: str) -> str:
    """
    Replaces illegal characters in NTFS file systems with
    appropriate substitutes
    :param string: Te string in which to replace the illegal characters
    :return: The sanitized string
    """
    illegal_characters = {
        "/": "ǁ",
        "\\": "ǁ",
        "?": "‽",
        "<": "←",
        ">": "→",
        ":": ";",
        "*": "∗",
        "|": "ǁ",
        "\"": "“"
    }
    for illegal_character, replacement in illegal_characters.items():
        string = string.replace(illegal_character, replacement)
    return string


def create_file(path: str):
    """
    Creates an empty file
    :param path: The path to the file
    :return: None
    """
    if not os.path.isfile(path):
        with open(path, "w") as f:
            f.write("")


def get_ext(filename: str) -> Optional[str]:
    """
    Gets the file extension of a file
    :param filename: The filename for which to get the file extension
    :return: The file extension or None if the file has no extension
    """
    try:
        return filename.rsplit(".", 1)[1]
    except IndexError:
        return None



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

import os
import logging
import pkg_resources
from typing import Type, Dict, Any, Callable, List, Optional
from bokkichat.settings.impl.TelegramBotSettings import TelegramBotSettings
from bokkichat.connection.impl.TelegramBotConnection import \
    TelegramBotConnection


class Config:
    """
    Class that keeps track of configuration data
    The class attributes should only be called after running load_config
    """

    @classmethod
    def load_config(cls, root_path: str, module_name: str, sentry_dsn: str):
        """
        Loads the configuration from environment variables
        :param root_path: The root path of the application
        :param module_name: The name of the project's module
        :param sentry_dsn: The sentry DSN used for error logging
        :return: None
        """
        cls.ensure_environment_variables_present()

        Config.LOGGING_PATH = os.environ.get(
            "LOGGING_PATH",
            os.path.join("/tmp", f"{module_name}.log")
        )
        Config.DEBUG_LOGGING_PATH = os.environ.get(
            "DEBUG_LOGGING_PATH",
            os.path.join("/tmp", f"{module_name}-debug.log")
        )
        verbosity_name = os.environ.get("VERBOSITY", "debug").lower()
        Config.VERBOSITY = {
            "error": logging.ERROR,
            "warning": logging.WARNING,
            "info": logging.INFO,
            "debug": logging.DEBUG
        }.get(verbosity_name, logging.DEBUG)

        Config.SENTRY_DSN = sentry_dsn
        Config.VERSION = \
            pkg_resources.get_distribution(module_name).version
        Config.FLASK_SECRET = os.environ["FLASK_SECRET"]
        Config.TESTING = os.environ.get("FLASK_TESTING") == "1"
        Config.BEHIND_PROXY = os.environ.get("BEHIND_PROXY") == "1"
        Config.HTTP_PORT = int(os.environ.get("HTTP_PORT", "80"))
        Config.DOMAIN_NAME = os.environ.get("DOMAIN_NAME", "localhost")

        if Config.TESTING:
            Config.DB_MODE = "sqlite"
        else:
            Config.DB_MODE = os.environ["DB_MODE"].lower()
        if Config.DB_MODE == "sqlite":
            sqlite_path = os.environ.get(
                "SQLITE_PATH",
                os.path.join("/tmp", f"{module_name}.db")
            )
            Config.DB_URI = "sqlite:///" + sqlite_path
        else:
            base = Config.DB_MODE.upper()
            db_keyword = "DATABASE"
            if base == "POSTGRESQL":
                base = "POSTGRES"
                db_keyword = "DB"
            base += "_"
            db_host = os.environ[base + "HOST"]
            db_port = os.environ[base + "PORT"]
            db_user = os.environ[base + "USER"]
            db_password = os.environ[base + "PASSWORD"]
            db_database = os.environ[base + db_keyword]
            Config.DB_URI = f"{Config.DB_MODE}://{db_user}:{db_password}@"\
                            f"{db_host}:{db_port}/{db_database}"

        Config.RECAPTCHA_SITE_KEY = os.environ["RECAPTCHA_SITE_KEY"]
        Config.RECAPTCHA_SECRET_KEY = os.environ["RECAPTCHA_SECRET_KEY"]

        Config.SMTP_HOST = os.environ["SMTP_HOST"]
        Config.SMTP_PORT = int(os.environ["SMTP_PORT"])
        Config.SMTP_ADDRESS = os.environ["SMTP_ADDRESS"]
        Config.SMTP_PASSWORD = os.environ["SMTP_PASSWORD"]
        Config.TELEGRAM_API_KEY = os.environ["TELEGRAM_API_KEY"]
        Config.TELEGRAM_WHOAMI = os.environ.get("TELEGRAM_WHOAMI", "1") == "1"

        cls._load_extras(Config)

        for required_template in cls.REQUIRED_TEMPLATES.values():
            path = os.path.join(root_path, "templates", required_template)
            if not os.path.isfile(path):
                print(f"Missing template file {path}")
                exit(1)

    @classmethod
    def _load_extras(cls, parent: Type["Config"]):
        """
        This method can be used to add attributes in subclasses as well as
        change attributes in the base Config class
        :param parent: The base Config class, used to chage attributes
        :return: None
        """
        pass

    @classmethod
    def environment_variables(cls) -> Dict[str, List[str]]:
        """
        Specifies required and optional environment variables
        :return: The specified environment variables in two lists in
                 a dictionary, grouped by whether the variables are
                 required or optional
        """
        required = [
            "FLASK_SECRET",
            "DB_MODE",
            "RECAPTCHA_SITE_KEY",
            "RECAPTCHA_SECRET_KEY",
            "SMTP_HOST",
            "SMTP_PORT",
            "SMTP_ADDRESS",
            "SMTP_PASSWORD",
            "TELEGRAM_API_KEY"
        ]
        optional = [
            "LOGGING_PATH",
            "DEBUG_LOGGING_PATH",
            "FLASK_TESTING",
            "DOMAIN_NAME",
            "HTTP_PORT",
            "BEHIND_PROXY",
            "TELEGRAM_WHOAMI"
        ]

        db_mode = os.environ.get("DB_MODE")
        if db_mode == "sqlite":
            optional.append("SQLITE_PATH")
        elif db_mode is not None:
            db_mode = db_mode.upper()
            required.append(f"{db_mode}_HOST")
            required.append(f"{db_mode}_PORT")
            required.append(f"{db_mode}_USER")
            required.append(f"{db_mode}_PASSWORD")
            required.append(f"{db_mode}_DATABASE")

            if db_mode == "POSTGRESQL":
                required += [
                    "POSTGRES_USER",
                    "POSTGRES_PASSWORD",
                    "POSTGRES_DB"
                ]

        return {
            "required": required,
            "optional": optional
        }

    @classmethod
    def ensure_environment_variables_present(cls):
        """
        Makes sure that all required environment variables have been set.
        If this is not the case, the app will exit.
        :return: None
        """
        for env_name in cls.environment_variables()["required"]:
            if env_name not in os.environ:
                print(f"Missing environment variable: {env_name}")
                exit(1)

    @classmethod
    def dump_env_variables(cls, path: Optional[str] = None):
        """
        Dumps all environment variables used by this application to a file
        :param path: The path to the file to which to dump the content.
                     If this is None, the file contents will be printed.
        :return: None
        """
        envs = ""
        all_env_names = cls.environment_variables()["required"] + \
            cls.environment_variables()["optional"]
        for env_name in all_env_names:
            value = os.environ.get(env_name)
            if value is not None:
                envs += f"{env_name}={value}\n"

        if path is not None:
            with open(path, "w") as f:
                f.write(envs)
        else:
            print(envs)

    @classmethod
    def base_url(cls) -> str:
        """
        :return: The base URL of the website
        """
        if cls.BEHIND_PROXY:
            return f"https://{cls.DOMAIN_NAME}"
        else:
            return f"http://{cls.DOMAIN_NAME}:{cls.HTTP_PORT}"

    @classmethod
    def initialize_telegram(cls):
        """
        Initializes the telegram bot connection
        :return: None
        """
        Config.TELEGRAM_BOT_CONNECTION = TelegramBotConnection(
            TelegramBotSettings(Config.TELEGRAM_API_KEY)
        )

    VERSION: str
    """
    The current version of the application
    """

    SENTRY_DSN: str
    """
    The sentry DSN used for error logging
    """

    VERBOSITY: int
    """
    The verbosity level of the logging when printing to the console
    """

    FLASK_SECRET: str
    """
    The flask secret key
    """

    TESTING: bool
    """
    Whether or not testing is enabled
    """

    LOGGING_PATH: str
    """
    The path to the logging file
    """

    DEBUG_LOGGING_PATH: str
    """
    The path to the debug logging path
    """

    WARNING_LOGGING_PATH: str
    """
    The path to the logging path for WARNING messages
    """

    HTTP_PORT: int
    """
    The port to use when serving the flask application
    """

    DOMAIN_NAME: str
    """
    The domain name of the website
    """

    DB_MODE: str
    """
    The database mode (for example 'sqlite' or 'mysql')
    """

    DB_URI: str
    """
    The database URI to use for database operations
    """

    RECAPTCHA_SITE_KEY: str
    """
    Google ReCaptcha site key for bot detection
    """

    RECAPTCHA_SECRET_KEY: str
    """
    Google ReCaptcha secret key for bot detection
    """

    SMTP_HOST: str
    """
    The SMPT Host used for sending emails
    """

    SMTP_PORT: int
    """
    The SMPT Port used for sending emails
    """

    SMTP_ADDRESS: str
    """
    The SMPT Address used for sending emails
    """
    SMTP_PASSWORD: str
    """
    The SMPT Password used for sending emails
    """

    TELEGRAM_API_KEY: str
    """
    API key for a telegram bot
    """

    TELEGRAM_BOT_CONNECTION: TelegramBotConnection
    """
    Telegram bot connection
    """

    TELEGRAM_WHOAMI: bool
    """
    Whether or not the telegram WHOAMI background thread will be started
    """

    BEHIND_PROXY: bool
    """
    Whether or not the site is being served by a reverse proxy like nginx.
    """

    MIN_USERNAME_LENGTH: int = 1
    """
    The Minimum length for usernames
    """

    MAX_USERNAME_LENGTH: int = 12
    """
    The maximum length of usernames
    """

    MAX_API_KEY_AGE: int = 2592000  # 30 days
    """
    The maximum age for API keys
    """

    API_VERSION: str = "1"
    """
    The API Version
    """

    REQUIRED_TEMPLATES: Dict[str, str] = {
        "index": "static/index.html",
        "about": "static/about.html",
        "privacy": "static/privacy.html",
        "error_page": "static/error_page.html",
        "registration_email": "email/registration.html",
        "forgot_password_email": "email/forgot_password.html",
        "forgot": "user_management/forgot.html",
        "login": "user_management/login.html",
        "profile": "user_management/profile.html",
        "register": "user_management/register.html"
    }
    """
    Specifies required template files
    """

    STRINGS: Dict[str, str] = {
        "401_message": "You are not logged in",
        "500_message": "The server encountered an internal error and "
                       "was unable to complete your request. "
                       "Either the server is overloaded or there "
                       "is an error in the application.",
        "user_does_not_exist": "User does not exist",
        "user_already_logged_in": "User already logged in",
        "user_already_confirmed": "User already confirmed",
        "user_is_not_confirmed": "User is not confirmed",
        "invalid_password": "Invalid Password",
        "logged_in": "Logged in successfully",
        "logged_out": "Logged out",
        "username_length": "Username must be between {} and {} characters "
                           "long",
        "passwords_do_not_match": "Passwords do not match",
        "email_already_in_use": "Email already in use",
        "username_already_exists": "Username already exists",
        "recaptcha_incorrect": "ReCaptcha not solved correctly",
        "registration_successful": "Registered Successfully. Check your email "
                                   "inbox for confirmation",
        "registration_email_title": "Registration",
        "confirmation_key_invalid": "Confirmation key invalid",
        "user_confirmed_successfully": "User confirmed successfully",
        "password_reset_email_title": "Password Reset",
        "password_was_reset": "Password was reset successfully",
        "password_changed": "Password changed successfully",
        "user_was_deleted": "User was deleted",
        "telegram_chat_id_set": "Telegram Chat ID was set"
    }
    """
    Dictionary that defines various strings used in the application.
    Makes it easier to use custom phrases.
    """

    TEMPLATE_EXTRAS: Dict[str, Callable[[], Dict[str, Any]]] = {
        "index": lambda: {},
        "about": lambda: {},
        "privacy": lambda: {},
        "login": lambda: {},
        "register": lambda: {},
        "forgot": lambda: {},
        "profile": lambda: {},
        "registration_email": lambda: {},
        "forgot_email": lambda: {}
    }
    """
    This can be used to provide the template rendering engine additional
    parameters, which may be necessary when adding UI elements.
    This is done with functions that don't expect any input and
    return a dictionary of keys and values to be passed to the template
    rendering engine
    """



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from enum import Enum
from typing import Dict, Any
from puffotter.flask.base import db


class ModelMixin:
    """
    A mixin class that specifies a couple of methods all database
    models should implement
    """

    id = db.Column(
        db.Integer, primary_key=True, nullable=False, autoincrement=True
    )
    """
    The ID is the primary key of the table and increments automatically
    """

    def __json__(self, include_children: bool = False) -> Dict[str, Any]:
        """
        Generates a dictionary containing the information of this model
        :param include_children: Specifies if children data models will be
                                 included or if they're limited to IDs
        :return: A dictionary representing the model's values
        """
        # TODO Make this automated.
        # repr and str both depend on this being implemented correctly.
        raise NotImplementedError()  # pragma: no cover

    def __str__(self) -> str:
        """
        :return: The string representation of this object
        """
        data = self.__json__()
        _id = data.pop("id")
        return "{}:{} <{}>".format(self.__class__.__name__, _id, str(data))

    def __repr__(self) -> str:
        """
        :return: A string with which the object may be generated
        """
        params = ""
        json_repr = self.__json__()

        enums = {}
        for key in json_repr:
            attr = getattr(self, key)
            if isinstance(attr, Enum):
                enum_cls = attr.__class__.__name__
                enum_val = attr.name
                enums[key] = "{}.{}".format(enum_cls, enum_val)

        for key, val in self.__json__().items():
            repr_arg = enums.get(key, repr(val))
            params += "{}={}, ".format(key, repr_arg)

        params = params.rsplit(",", 1)[0]

        return "{}({})".format(self.__class__.__name__, params)

    def __eq__(self, other: Any) -> bool:
        """
        Checks the model object for equality with another object
        :param other: The other object
        :return: True if the objects are equal, False otherwise
        """
        if "__json__" in dir(other):
            return other.__json__() == self.__json__()
        else:
            return False  # pragma: no cover

    def __hash__(self) -> int:
        """
        Creates a hash so that the model objects can be used as keys
        :return: None
        """
        return hash(repr(self))



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from unittest.mock import patch
from puffotter.flask.test.TestFramework import _TestFramework
from puffotter.flask.db.User import User
from puffotter.flask.Config import Config


class TestRegisterRoute(_TestFramework):
    """
    Class that tests registration features
    """

    def test_page_get(self):
        """
        Tests getting the page
        :return: None
        """
        resp = self.client.get("/register")
        self.assertTrue(b"<!--user_management/register.html-->" in resp.data)

    def test_registering_user(self):
        """
        Tests registering a new user
        :return: None
        """
        self.assertEqual(len(User.query.all()), 0)
        with self.client:

            with patch("puffotter.flask.routes.user_management.send_email") \
                    as m:
                self.assertEqual(0, m.call_count)
                resp = self.client.post(
                    "/register",
                    follow_redirects=True,
                    data={
                        "username": "testuser",
                        "password": "testpass",
                        "password-repeat": "testpass",
                        "email": "test@example.com",
                        "g-recaptcha-response": ""
                    }
                )
                self.assertEqual(1, m.call_count)

            registration = self.config.STRINGS["registration_successful"]
            self.assertTrue(registration.encode("utf-8") in resp.data)
            self.assertTrue(b"<!--static/index.html-->" in resp.data)
            self.assertEqual(len(User.query.all()), 1)

            user = User.query.filter_by(username="testuser").first()
            self.assertEqual(user.email, "test@example.com")
            self.assertFalse(user.confirmed)

    def test_invalid_registrations(self):
        """
        Tests registering using invalid parameters
        :return: None
        """
        user, _, _ = self.generate_sample_user()
        base = {
            "username": "testuser",
            "password": "testpass",
            "password-repeat": "testpass",
            "email": "test@example.com",
            "g-recaptcha-response": ""
        }
        self.assertEqual(len(User.query.all()), 1)
        with self.client:

            for params in [
                {"password": "testpasses",
                 "expected": "passwords_do_not_match"},
                {"username": "-" * (Config.MIN_USERNAME_LENGTH - 1),
                 "expected": "username_length"},
                {"username": "-" * (Config.MAX_USERNAME_LENGTH + 1),
                 "expected": "username_length"},
                {"username": user.username,
                 "expected": "username_already_exists"},
                {"email": user.email,
                 "expected": "email_already_in_use"}
            ]:
                data = dict(base)
                data.update(params)

                with patch("puffotter.flask.routes.user_management."
                           "send_email")as m:
                    self.assertEqual(0, m.call_count)
                    resp = self.client.post(
                        "/register",
                        follow_redirects=True,
                        data=data
                    )
                    self.assertEqual(0, m.call_count)

                expected = self.config.STRINGS[params["expected"]]
                if params["expected"] == "username_length":
                    expected = expected[0:20]

                registration = self.config.STRINGS["registration_successful"]
                self.assertFalse(registration.encode("utf-8") in resp.data)
                self.assertTrue(
                    b"<!--user_management/register.html-->" in resp.data
                )
                self.assertTrue(expected.encode("utf-8") in resp.data)
                self.assertEqual(len(User.query.all()), 1)

    def test_invalid_recaptcha(self):
        """
        Tests that invalid ReCaptcha responses are handled correctly
        :return: None
        """
        with self.client:
            with patch("puffotter.flask.routes.user_management.send_email") \
                    as m:
                with patch("puffotter.flask.routes.user_management."
                           "verify_recaptcha", lambda x, y, z: False):
                    self.assertEqual(0, m.call_count)
                    resp = self.client.post(
                        "/register",
                        follow_redirects=True,
                        data={
                            "username": "testuser",
                            "password": "testpass",
                            "password-repeat": "testpass",
                            "email": "test@example.com",
                            "g-recaptcha-response": ""
                        }
                    )
                    self.assertEqual(0, m.call_count)

            registration = self.config.STRINGS["registration_successful"]
            recaptcha = self.config.STRINGS["recaptcha_incorrect"]
            self.assertFalse(registration.encode("utf-8") in resp.data)
            self.assertTrue(recaptcha.encode("utf-8") in resp.data)
            self.assertTrue(
                b"<!--user_management/register.html-->" in resp.data
            )
            self.assertEqual(len(User.query.all()), 0)

    def test_confirming(self):
        """
        Tests confirming a user
        :return: None
        """
        user, _, confirm_key = self.generate_sample_user(False)
        with self.client:
            resp = self.client.get("/confirm?user_id={}&confirm_key={}".format(
                user.id, confirm_key
            ), follow_redirects=True)
            confirmed = self.config.STRINGS["user_confirmed_successfully"]
            self.assertTrue(confirmed.encode("utf-8") in resp.data)
            self.assertTrue(b"<!--static/index.html-->" in resp.data)
            self.assertTrue(user.confirmed)

    def test_invalid_confirm(self):
        """
        Tests invalid confirmations
        :return: None
        """
        user, _, confirm_key = self.generate_sample_user(False)
        user2, _, confirm_key2 = self.generate_sample_user()
        with self.client:
            for params in [
                {"user_id": "100", "confirm_key": confirm_key,
                 "expected": "user_does_not_exist"},
                {"user_id": user.id, "confirm_key": "AAA",
                 "expected": "confirmation_key_invalid"},
                {"user_id": user2.id, "confirm_key": confirm_key2,
                 "expected": "user_already_confirmed"}
            ]:
                resp = self.client.get(
                    "/confirm?user_id={}&confirm_key={}".format(
                        params["user_id"], params["confirm_key"]
                    ), follow_redirects=True)
                expected = self.config.STRINGS[params["expected"]]
                confirmed = self.config.STRINGS["user_confirmed_successfully"]

                self.assertFalse(confirmed.encode("utf-8") in resp.data)
                self.assertTrue(b"<!--static/index.html-->" in resp.data)
                self.assertTrue(expected.encode("utf-8") in resp.data)
                self.assertFalse(user.confirmed)



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""

from puffotter.flask.test.TestFramework import _TestFramework


class TestStaticRoutes(_TestFramework):
    """
    Class that tests static pages
    """

    def test_get_index(self):
        """
        Tests getting the index page
        :return: None
        """
        resp = self.client.get("/")
        self.assertTrue(b"<!--static/index.html-->" in resp.data)

    def test_get_about(self):
        """
        Tests getting the about page
        :return: None
        """
        resp = self.client.get("/about")
        self.assertTrue(b"<!--static/about.html-->" in resp.data)

    def test_get_privacy(self):
        """
        Tests getting the privacy page
        :return: None
        """
        resp = self.client.get("/privacy")
        self.assertTrue(b"<!--static/privacy.html-->" in resp.data)



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""


class ApiException(Exception):
    """
    Api raised when an API-related exception occurs
    """

    def __init__(self, reason: str, status_code: int):
        """
        Initializes the exception
        :param reason: The reason the API Exception was raised
        :param status_code: The status code associated with the exception
        """
        super().__init__(reason)
        self.reason = reason
        self.status_code = status_code



"""LICENSE
Copyright 2019 Hermann Krumrey <hermann@krumreyh.com>

This file is part of puffotter.

puffotter is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

puffotter is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with puffotter.  If not, see <http://www.gnu.org/licenses/>.
LICENSE"""


# noinspection PyPackageRequirements
from telegram.error import Conflict
from puffotter.flask.base import app
from puffotter.flask.Config import Config
from bokkichat.entities.message.TextMessage import TextMessage


def telegram_whoami():
    """
    Specifies the background behaviour of the telegram bot
    By default, the bot listens to /whoami messages
    and answers with the telegram chat ID
    :return: None
    """
    telegram = Config.TELEGRAM_BOT_CONNECTION

    def handler(_, msg):
        if msg.is_text():
            msg: TextMessage = msg

            if msg.body == "/whoami":
                sender = telegram.address
                receiver = msg.sender
                telegram.send(
                    TextMessage(sender, receiver, receiver.address))

    try:
        telegram.loop(handler)
    except Conflict:
        app.logger.warning("It seems that two instances of the telegram "
                           "bot are running")
